/*
 * Copyright (c) 2010 Douglas Gilbert.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*****************************************************************
 * setbits.c
 * 
 * Utility for setting a GPIO line on a AT91SAM9G20/45 in linux.
 * Uses sysfs interface. The target hardware is a FoxG20 and this utility
 * mimics the actions of a utility of the same name for the FoxLX board.
 *
 ****************************************************/

#define _XOPEN_SOURCE 500

#include <stdio.h>     
#include <stdlib.h>     
#include <ctype.h>     
#include <string.h>     
#include <unistd.h>    
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>


static const char * version_str = "1.01 20100218";

#define EXPORT_FILE "/sys/class/gpio/export"
#define UNEXPORT_FILE "/sys/class/gpio/unexport"
#define GPIO_BASE_FILE "/sys/class/gpio/gpio"

static int verbose = 0;


static void
usage(void)
{
    fprintf(stderr, "Usage: "
            "setbits [-b <bn>] [-h] -p <port> [-s 0|1] [-u] "
            "[-v] [-V]\n"
            "  where:\n"
            "    -b <bn>      bit number within a port (0 to 31)\n"
            "    -h           print usage message\n"
            "    -p <port>    port ('a' to 'e') or gpio kernel line number "
            "(32 or more)\n"
            "    -s 0|1       state to set, 0 for low, 1 for high\n"
            "                 (def: set to input, usually with weak "
            "pullup)\n"
            "    -u           unexport gpio line prior to setting bit\n"
            "    -v           increase verbosity (multiple times for more)\n"
            "    -V           print version string then exit\n\n"
            "Set GPIO line state on AT91SAM9G20/45. For FoxG20 and similar "
            "to\na FoxLX utility of the same name.\n");
}


int
main(int argc, char ** argv)
{
    int opt, k, exp_fd, unexp_fd, dir_fd;
    int bn = -1;
    int knum = -1;
    int state = -1;
    int unexport = 0;
    int exported = 0;
    int ret = -1;
    char b[128];
    char ch;
    char bank = '\0';

    while ((opt = getopt(argc, argv, "b:hp:s:uvV")) != -1) {
        switch (opt) {
        case 'b':
            k = atoi(optarg);
            if ((k < 0) || (k > 31)) {
                fprintf(stderr, "'-b' expects a bit number from 0 to 31\n");
                exit(EXIT_FAILURE);
            }
            bn = k;
            break;
        case 'h':
            usage();
            exit(EXIT_SUCCESS);
            break;
        case 'p':
            if (isalpha(*optarg)) {
                ch = toupper(*optarg);
                if ((ch >= 'A') && (ch <= 'E'))
                    bank = ch;
                else {
                    fprintf(stderr, "'-p' expects a letter ('A' to 'E') or "
                            "a number\n");
                    exit(EXIT_FAILURE);
                }
            } else if (isdigit(*optarg)) {
                k = atoi(optarg);
                if ((k < 32) || (k > 511)) {
                    fprintf(stderr, "'-p' expects a letter or a number "
                            "32 or greater\n");
                    exit(EXIT_FAILURE);
                }
                knum = k;
            } else {
                fprintf(stderr, "'-p' expects a letter ('A' to 'E') or "
                        "a number\n");
                exit(EXIT_FAILURE);
            }
            break;
        case 's':
            state = atoi(optarg);
            if ((state < 0) || (state > 1)) {
                fprintf(stderr, "'-s' expects '0' or '1'\n");
                exit(EXIT_FAILURE);
            }
            break;
        case 'u':
            ++unexport;
            break;
        case 'v':
            ++verbose;
            break;
        case 'V':
            printf("%s\n", version_str);
            exit(EXIT_SUCCESS);
        default: /* '?' */
            usage();
            exit(EXIT_FAILURE);
        }
    }
    if (optind < argc) {
        if (optind < argc) {
            for (; optind < argc; ++optind)
                fprintf(stderr, "Unexpected extra argument: %s\n",
                        argv[optind]);
            usage();
            exit(EXIT_FAILURE);
        }
    }

    if (! ((knum >= 0) || ((bank >= 'A') && (bn >= 0)))) {
        fprintf(stderr, "Expect either '-p <knum>' or ('-p <letter>' and "
                "'-b <bn>')\n");
        usage();
        exit(EXIT_FAILURE);
    }
    if (knum < 0) {
        knum = (1 + bank - 'A') * 32 + bn;
        if (verbose)
            fprintf(stderr, "%c%d becomes kernel pin number %d\n", bank, 
                    bn, knum);
    }

    exp_fd = -1;
    unexp_fd = -1;
    dir_fd = -1;
    exp_fd = open(EXPORT_FILE, O_WRONLY);
    if (exp_fd < 0) {
        perror(EXPORT_FILE);
        goto bad;
    }
    unexp_fd = open(UNEXPORT_FILE, O_WRONLY);
    if (unexp_fd < 0) {
        perror(UNEXPORT_FILE);
        goto bad;
    }

    if (unexport) {
        snprintf(b, sizeof(b), "%d", knum);
        if (pwrite(unexp_fd, b, strlen(b), 0) < 0) {
            fprintf(stderr, "Unable to unexport knum=%d: %s\n", knum,
                    strerror(errno));
            fprintf(stderr, "continue ...\n");
        }
    }
    snprintf(b, sizeof(b), "%d", knum);
    if (pwrite(exp_fd, b, strlen(b), 0) < 0) {
        fprintf(stderr, "Unable to export knum=%d (already in use?): %s\n",
                knum, strerror(errno));
        goto bad;
    }
    exported = 1;
    snprintf(b, sizeof(b), "%s%d/direction", GPIO_BASE_FILE, knum);
    dir_fd = open(b, O_WRONLY);
    if (dir_fd < 0) {
        fprintf(stderr, "Open %s: %s\n", b, strerror(errno));
        goto bad;
    }
    if (state < 0) {
        if (pwrite(dir_fd, "in", 2, 0) < 0) {
            fprintf(stderr, "Unable to write 'in' to dir_fd: %s\n",
                    strerror(errno));
            goto bad;
        }
    } else if (0 == state) {
        if (pwrite(dir_fd, "low", 3, 0) < 0) {
            fprintf(stderr, "Unable to write 'low' to dir_fd: %s\n",
                    strerror(errno));
            goto bad;
        }
    } else {
        if (pwrite(dir_fd, "high", 4, 0) < 0) {
            fprintf(stderr, "Unable to write 'high' to dir_fd: %s\n",
                    strerror(errno));
            goto bad;
        }
    }
    ret = 0;

bad:
    if (dir_fd >= 0)
        close(dir_fd);
    if (unexp_fd >= 0) {
        if (exported) {
            snprintf(b, sizeof(b), "%d", knum);
            pwrite(unexp_fd, b, strlen(b), 0);
        }
        close(unexp_fd);
    }
    if (exp_fd >= 0)
        close(exp_fd);
    if (ret)
        exit(EXIT_FAILURE);
    else
        return 0;
}
